import logging

import pandas as pd

from .exceptions import OepClientSideException
from .utils import dataframe_to_records


class AdvancedApiSession:
    """Context for advanced api session (close connection on exit)"""

    def __init__(self, oepclient):
        """
        Args:

            oepclient(OEPClient)
        """
        self.oepclient = oepclient
        self.api_url = self.oepclient.api_url + "advanced/"
        self.connection_id = None
        self.cursor_id = None

    def _command(self, path, query=None, verb="POST", expected_status=200):
        url = self.api_url + path
        jsondata = {}
        if self.connection_id:
            jsondata["connection_id"] = self.connection_id
        if self.cursor_id:
            jsondata["cursor_id"] = self.cursor_id
        if query:
            jsondata["query"] = query
        logging.debug(jsondata)
        res = self.oepclient._request(verb, url, expected_status, jsondata)
        return res

    def __enter__(self):
        self.connection_id = self._command("connection/open")["content"][
            "connection_id"
        ]
        self.cursor_id = self._command("cursor/open")["content"]["cursor_id"]
        logging.debug("Started connection: %s", self.connection_id)
        return self

    def __exit__(self, _exc_type, exc_val, _exc_tb):
        # rollback on error, otherwise commit
        if exc_val:
            logging.error(exc_val)
            self._command("connection/rollback")
        else:
            self._command("connection/commit")
        if self.cursor_id:
            self._command("cursor/close")
            logging.debug("Closed cursor: %s", self.cursor_id)
            self.cursor_id = None
        if self.connection_id:
            self._command("connection/close")
            logging.debug("Closed connection: %s", self.connection_id)
            self.connection_id = None

    def _get_query(self, table, **kwargs):
        query = {"table": table}
        query.update(kwargs)
        return query

    def insert_into_table(self, table, data):
        """Insert records into table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
            data(list): list of records(dict: column_name -> value)
        """
        if isinstance(data, pd.DataFrame):
            data = dataframe_to_records(data)

        if not isinstance(data, (list, tuple)) or (
            data and not isinstance(data[0], dict)
        ):
            raise OepClientSideException(
                "data must be list or tuple of record dictionaries"
            )

        query = self._get_query(table, values=data)
        return self._command("insert", query)

    def select_from_table(self, table):
        """Select all rows from table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

        Returns:
            list of records(dict: column_name -> value)
        """

        query = self._get_query(table)
        self._command("search", query)
        return self._command("cursor/fetchall")

    def delete_from_table(self, table):
        """Delete all rows from table (without dropping it).

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
        """

        query = self._get_query(table)
        return self._command("delete", query)



class OepApiException(Exception):
    pass


class OepServerSideException(OepApiException):
    pass


class OepClientSideException(OepApiException):
    pass


class OepAuthenticationException(OepClientSideException):
    pass


class OepTableNotFoundException(OepClientSideException):
    def __init__(self, _msg=None):
        # the API falsely returns message: {
        # 'detail': 'You do not have permission to perform this action}
        # but this is only because table  does not exist
        # TODO: create better error message on server side!
        super().__init__(_msg or "Table does not exist OR you don't have permission")


class OepTableAlreadyExistsException(OepClientSideException):
    pass



"""Command line script for OepClient"""

import json
import logging
import os
import sys

import click

from oep_client import __version__
from oep_client.exceptions import OepApiException
from oep_client.oep_client import (
    DEFAULT_API_VERSION,
    DEFAULT_BATCH_SIZE,
    DEFAULT_HOST,
    DEFAULT_INSERT_RETRIES,
    DEFAULT_PROTOCOL,
    TOKEN_ENV_VAR,
    OepClient,
)
from oep_client.test import TestRoundtrip
from oep_client.utils import (
    dataframe_to_records,
    get_schema_definition_from_metadata,
    read_dataframe,
    read_metadata_json,
    records_to_dataframe,
    write_dataframe,
    write_json,
)

PROG_NAME = "oep-client"
LOGGING_DATE_FMT = "%Y-%m-%d %H:%M:%S"
LOGGING_FMT = "[%(asctime)s.%(msecs)03d %(levelname)7s] %(message)s"
DEFAULT_ENCODING = "utf-8"


@click.group()
@click.pass_context
@click.version_option(__version__, prog_name=PROG_NAME)
@click.option(
    "--loglevel",
    "-l",
    default="info",
    type=click.Choice(["debug", "info", "warning", "error"]),
)
@click.option("--token", "-t")
@click.option("--protocol", default=DEFAULT_PROTOCOL)
@click.option("--host", default=DEFAULT_HOST)
@click.option("--api-version", default=DEFAULT_API_VERSION)
@click.option("--batch-size", "-b", default=DEFAULT_BATCH_SIZE)
@click.option("--insert-retries", default=DEFAULT_INSERT_RETRIES)
def main(
    ctx,
    loglevel,
    token,
    protocol,
    host,
    api_version,
    batch_size,
    insert_retries,
):
    loglevel = getattr(logging, loglevel.upper())
    logging.basicConfig(format=LOGGING_FMT, datefmt=LOGGING_DATE_FMT, level=loglevel)
    ctx.ensure_object(dict)

    token = token or os.environ.get(TOKEN_ENV_VAR)

    ctx.obj["client"] = OepClient(
        token=token,
        protocol=protocol,
        host=host,
        api_version=api_version,
        batch_size=batch_size,
        insert_retries=insert_retries,
    )


@main.command("create")
@click.pass_context
@click.argument("table")
@click.argument("metadata_file", type=click.Path())
@click.option("--encoding", "-e", default=DEFAULT_ENCODING)
@click.option("--upload-metadata", "-m", is_flag=True)
@click.option("--is-sandbox", is_flag=True)
def create_table(
    ctx, table, metadata_file, encoding, upload_metadata, is_sandbox: bool = False
):
    metadata = read_metadata_json(metadata_file, encoding)
    definition = get_schema_definition_from_metadata(metadata)
    client = ctx.obj["client"]
    client.create_table(table, definition, is_sandbox=is_sandbox)
    # automatically upload metadata
    if upload_metadata:
        client.set_metadata(table, metadata)
    logging.info("OK")


@main.command("drop")
@click.pass_context
@click.argument("table")
def drop_table(ctx, table):
    client = ctx.obj["client"]
    client.drop_table(table)
    logging.info("OK")


@main.command("insert")
@click.pass_context
@click.argument("table")
@click.argument("data_file", type=click.Path())
@click.option("--encoding", "-e", default=DEFAULT_ENCODING)
@click.option("--sheet", "-s", default=None)
@click.option("--delimiter", "-d", default=",")
def insert_into_table(ctx, table, data_file, encoding, sheet, delimiter):
    df = read_dataframe(data_file, encoding=encoding, sheet=sheet, delimiter=delimiter)
    data = dataframe_to_records(df)
    client = ctx.obj["client"]
    client.insert_into_table(table, data)
    logging.info("OK")


@main.command("select")
@click.pass_context
@click.argument("table")
@click.argument("data_file", type=click.Path(exists=False), required=False)
@click.option("--where", "-w", multiple=True)
@click.option("--sheet", "-s", default=None)
@click.option("--delimiter", "-d", default=",")
def select_from_table(ctx, table, data_file, where, sheet, delimiter):
    client = ctx.obj["client"]
    data = client.select_from_table(table, where=where)
    if data_file:
        df = records_to_dataframe(data)
        write_dataframe(df, data_file, sheet=sheet, delimiter=delimiter)
    else:
        # print to stdout
        datas = json.dumps(data, ensure_ascii=False, indent=2)
        datab = datas.encode()
        sys.stdout.buffer.write(datab)

    logging.info("OK")


@main.group("metadata")
def metadata():
    pass


@metadata.command("get")
@click.pass_context
@click.argument("table")
@click.argument("metadata_file", type=click.Path(exists=False), required=False)
@click.option("--encoding", "-e", default=DEFAULT_ENCODING)
def get_metadata(ctx, table, metadata_file, encoding):
    client = ctx.obj["client"]
    metadata = client.get_metadata(table)

    if metadata_file:
        write_json(metadata, metadata_file, encoding=encoding)
    else:
        # print to stdout
        datas = json.dumps(metadata, ensure_ascii=False, indent=2)
        datab = datas.encode()
        sys.stdout.buffer.write(datab)

    logging.info("OK")


@metadata.command("set")
@click.pass_context
@click.argument("table")
@click.argument("metadata_file", type=click.Path(exists=False))
@click.option("--encoding", "-e", default=DEFAULT_ENCODING)
def set_metadata(ctx, table, metadata_file, encoding):
    metadata = read_metadata_json(metadata_file, encoding)
    client = ctx.obj["client"]
    client.set_metadata(table, metadata)
    logging.info("OK")


@main.command("test")
@click.pass_context
def test_roundtrip(ctx):
    client = ctx.obj["client"]
    TestRoundtrip().test_roundtrip(client=client)
    logging.info("OK")


@main.command("publish")
@click.pass_context
@click.argument("table")
@click.argument("topic")
def publish_table(ctx, table, topic):
    client = ctx.obj["client"]
    client.publish_table(table, topic)
    logging.info("OK")


@main.command("unpublish")
@click.pass_context
@click.argument("table")
def unpublish_table(ctx, table):
    client = ctx.obj["client"]
    client.unpublish_table(table)
    logging.info("OK")


@main.command("count")
@click.pass_context
@click.argument("table")
def count(ctx, table):
    client = ctx.obj["client"]
    n_rows = client.count_rows(table)
    logging.info("%d", n_rows)


@main.command("delete")
@click.pass_context
@click.argument("table")
def delete(ctx, table):
    client = ctx.obj["client"]
    client.delete_from_table(table)
    logging.info("OK")


@main.command("list")
@click.pass_context
def iter_tables(ctx):
    client = ctx.obj["client"]
    for item in client.iter_tables():
        print("%(table)s" % item)


if __name__ == "__main__":
    try:
        main(prog_name=PROG_NAME)
    except OepApiException as exc:
        logging.error("%s: %s", exc.__class__.__name__, str(exc))
        sys.exit(1)



"""Work in Progress - requires sqlalchemy and oedialect"""

import os


def get_sqlalchemy_table(oepclient, table):
    """
    Args:

        oepclient(OEPClient)
    """
    # NOTE: import inside of function because it's not mandatory
    import sqlalchemy as sa  # noqa

    os.environ["OEDIALECT_PROTOCOL"] = oepclient.protocol

    if oepclient.token:
        connection_string = "postgresql+oedialect://:%s@%s" % (
            oepclient.token,
            oepclient.host,
        )
    else:
        connection_string = "postgresql+oedialect://%s" % (oepclient.host,)

    engine = sa.create_engine(connection_string)
    metadata = sa.MetaData(bind=engine)

    parts = []
    for col in oepclient.get_table_definition(table)["columns"]:
        parts.append(sa.Column(col["name"]))

    tab = sa.Table(table, metadata, *parts)
    return tab



"""Small client implementation to use the oep API.

Example usage: create a table, insert data, retrieve data, delete table

cli = OepClient(token='xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')

table = 'my_awesome_test_table'
definition = {
    "columns": [
        {"name": "id", "data_type": "bigint", "is_nullable": False, "primary_key": True},
        {"name": "field1", "data_type": "varchar(128)", "is_nullable": False},
        {"name": "field2", "data_type": "integer", "is_nullable": True}
    ]
}
data =  [
    {'id': 1, 'field1': 'test', 'field2': 100},
    {'id': 2, 'field1': 'test2', 'field2': None}
]

cli.create_table(table, definition)
cli.insert_into_table(table, data)
return_data = cli.select_from_table(table)
cli.drop_table(table)

"""  # noqa

import functools
import logging
import math
import re

import click
import pandas as pd
import requests

from .advanced_api import AdvancedApiSession

# from .dialect import get_sqlalchemy_table
from .exceptions import (
    OepAuthenticationException,
    OepClientSideException,
    OepServerSideException,
    OepTableAlreadyExistsException,
    OepTableNotFoundException,
)
from .utils import dataframe_to_records, fix_table_definition

DEFAULT_HOST = "openenergyplatform.org"
DEFAULT_PROTOCOL = "https"
DEFAULT_API_VERSION = "v0"
DEFAULT_BATCH_SIZE = 5000
DEFAULT_INSERT_RETRIES = 10
TOKEN_ENV_VAR = "OEP_API_TOKEN"


def check_exception(pattern: str, exception):
    """create decorator for custom Exceptions."""

    def decorator(fun):
        @functools.wraps(fun)
        def _fun(*args, **kwargs):
            try:
                return fun(*args, **kwargs)
            except Exception as exc:
                msg = str(exc)
                if re.match(".*" + pattern, msg, re.IGNORECASE):
                    raise exception(msg)
                raise

        return _fun

    return decorator


class OepClient:
    """Small client implementation to use the oep API."""

    def __init__(
        self,
        token=None,
        protocol=DEFAULT_PROTOCOL,
        host=DEFAULT_HOST,
        api_version=DEFAULT_API_VERSION,
        batch_size=DEFAULT_BATCH_SIZE,
        insert_retries=DEFAULT_INSERT_RETRIES,
    ):
        """
        Args:
            token(str): your API token
            host(str, optional): default is "https". "http" may be used for
              local installations
            host(str, optional): host of the oep platform.
              default is "openenergyplatform.org"
            api_version(str, optional): currently only "v0"
            batch_size(int, optional): number of records that will be uploaded
              per batch.
               if 0 or None: do not use batches
            insert_retries(int, optional): number of insert_retries for insert
               on OepServerSideExceptions
        """
        self.headers = {"Authorization": "Token %s" % token} if token else {}
        self.api_url = "%s://%s/api/%s/" % (protocol, host, api_version)
        self.web_url = "%s://%s/database/" % (protocol, host)
        self.protocol = protocol
        self.host = host
        self.token = token
        self.batch_size = batch_size
        self.insert_retries = insert_retries

    def _get_table_api_url(self, table):
        """Return base api url for table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
        """

        url = self.api_url + "tables/%s/" % table
        # url = self.api_url + "tables/%s/" % table

        logging.debug("URL: %s", url)
        return url

    def get_web_url(self, table):
        """Return web url for data edit/view

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
        """
        url = self.web_url + "tables/%s" % table
        logging.debug("URL: %s", url)
        return url

    @check_exception("invalid token", OepAuthenticationException)
    def _request(
        self, method, url, expected_status, jsondata=None, params: dict | None = None
    ):
        """Send a request and perform basic check for results

        Args:
            method(str): http method, that will be passed on to `requests.request`
            url(str): request url
            expected_status(int): expected http status code.
                if result has a different code, an error will be raised
            jsondata(object, optional): payload that will be send as json
                in the request.
        Returns:
            result object from returned json data
        """
        res = requests.request(
            url=url, method=method, json=jsondata, headers=self.headers, params=params
        )
        logging.debug("%d %s %s", res.status_code, method, url)
        try:
            res_json = res.json()
        except Exception:
            # api should return json, but some actions don't,
            # and 500 errors obviously also don't
            res_json = {}
        if res.status_code >= 500:
            raise OepServerSideException(res_json)
        elif res.status_code >= 400:
            raise OepClientSideException(res_json)
        if res.status_code != expected_status:
            if "reason" in res_json:
                res_json = res_json["reason"]
            raise OepClientSideException(res_json)
        return res_json

    @check_exception("exists", OepTableAlreadyExistsException)
    def create_table(self, table, definition, is_sandbox: bool = False):
        """Create table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
            definition(object): column and constraint definitions
                according to the oep specifications

                Notes:
                * data_type should be understood by postgresql database
                * is_nullable: True or False
                * the first column should be the primary key
                  and a numeric column with the name `id` for full functionality om the platform

                Example:
                {
                    "columns": [
                        {"name": "id", "data_type": "bigint", "is_nullable": False, "primary_key": True},
                        {"name": "field1", "data_type": "varchar(128)", "is_nullable": False},
                        {"name": "field2", "data_type": "integer", "is_nullable": True}
                    ]
                }

        """  # noqa
        url = self._get_table_api_url(table=table)
        definition = fix_table_definition(definition)
        logging.debug(definition)
        if is_sandbox:
            # NOTE: do NOT send {"is_sandbox": False}, API checks only if
            # query param is_sandbox is set
            params = {"is_sandbox": True}
        else:
            params = {}

        self._request("PUT", url, 201, {"query": definition}, params=params)
        definition_final = self.get_table_definition(table=table)
        logging.debug(definition_final)

    # inconsistent message from server:
    # "do not have permission" when table does not exist
    @check_exception("do not have permission", OepTableNotFoundException)
    def drop_table(self, table):
        """Drop table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
        """
        url = self._get_table_api_url(table=table)
        return self._request("DELETE", url, 200)

    @check_exception("not found", OepTableNotFoundException)
    def select_from_table(self, table, where=None):
        """Select all rows from table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
            where(list, optional): filter criteria in form of field/operator/value,
                e.g. ["id>10"]

        Returns:
            list of records(dict: column_name -> value)
        """
        url = self._get_table_api_url(table=table) + "rows/"

        if where:
            # convert dict into url
            where = "&".join(f"where={w}" for w in where)
            url = f"{url}?{where}"

        res = self._request("GET", url, 200)
        return res

    # inconsistent message from server:
    # "do not have permission" when table does not exist
    @check_exception("do not have permission", OepTableNotFoundException)
    def insert_into_table(self, table, data, batch_size=None, method="api"):
        """Insert records into table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
            data(list): list of records(dict: column_name -> value)
            batch_size(int, optional): defaults to client's default batch size
            method(list, optional):
                * 'api': sent records via regular API
                * 'advanced' (default): sent records via advanced API
        """

        table_def = self.get_table_definition(table)
        column_names = [c["name"] for c in table_def["columns"]]

        if isinstance(data, pd.DataFrame):
            used_column_names = set(data.columns)
            data = dataframe_to_records(data)
        else:
            used_column_names = set()
            for row in data:
                used_column_names = used_column_names | set(row.keys())

        # FIXME: on oep server: columns are determined by keys in first row!
        # for now, we have to fix at least the first row
        if data and set(data[0]) < used_column_names:
            for c in used_column_names - set(data[0]):
                data[0][c] = None

        unknown_column_names = used_column_names - set(column_names)
        if unknown_column_names:
            raise OepClientSideException(
                "Columns not in table: %s", unknown_column_names
            )

        batch_size = batch_size or self.batch_size
        n_batches = math.ceil(len(data) / batch_size)
        data_batches = []
        for i in range(n_batches):
            i_from = i * batch_size
            i_to = (i + 1) * batch_size
            data_batches.append(data[i_from:i_to])

        n_items = 0
        with click.progressbar(data_batches) as data_parts:
            for i_item, data_part in enumerate(data_parts):
                logging.debug(
                    "Starting upload batch %d/%d (%d/%d)...",
                    i_item + 1,
                    n_batches,
                    n_items,
                    len(data),
                )
                try_number = 0
                success = False
                while try_number <= self.insert_retries:
                    try_number += 1
                    try:
                        if method == "api":
                            self._insert_into_table_api(table, data_part)
                        elif method == "advanced":
                            with AdvancedApiSession(self) as ses:
                                ses.insert_into_table(table, data_part)
                        else:
                            raise NotImplementedError(method)
                        # batch upload ok
                        success = True
                        break
                    except OepServerSideException:
                        if try_number <= self.insert_retries:
                            # FIXME ???
                            logging.debug("A server side error occurred. retrying...")

                if not success:
                    raise OepServerSideException()

                n_items += len(data_part)

        return self.count_rows(table=table)

    def _insert_into_table_api(self, table, data):
        """Insert records into table.

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore
            data(list): list of records(dict: column_name -> value)
        """
        if not data:
            logging.warning("no data")
            return {}

        url = self._get_table_api_url(table=table) + "rows/new"
        res = self._request("POST", url, 201, {"query": data})
        return res

    @check_exception("not found", OepTableNotFoundException)
    def get_table_definition(self, table):
        """Returns table info

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

        """
        url = self._get_table_api_url(table=table)
        res = self._request("GET", url, 200)
        definition = {
            "columns": [],
            "constraints": [],
        }

        # res["columns"] is dict
        # first: add constrains to field definitions / constraints
        for const in res["constraints"].values():
            const_type = const["constraint_type"]
            const_def = const["definition"]
            if const_type == "PRIMARY KEY":
                args = re.match(
                    r"^PRIMARY KEY \((?P<field>[^)]+)\)$", const_def
                ).groupdict()
                # NOTE currently only single field PK allowed
                res["columns"][args["field"]]["primary_key"] = True
            elif const_type == "FOREIGN KEY":
                args = re.match(
                    r"^FOREIGN KEY \((?P<field>[^)]+)\) REFERENCES (?P<ref_table>[^()]+)\((?P<ref_field>[^)]+)\)$",  # noqa
                    const_def,
                ).groupdict()
                # NOTE currently only single field PK allowed
                res["columns"][args["field"]]["foreign_key"] = [
                    {
                        "table": args["ref_table"],
                        "column": args["ref_field"],
                    }
                ]
            elif const_type == "UNIQUE":
                args = re.match(
                    r"^UNIQUE \((?P<fields>[^)]+)\)$", const_def
                ).groupdict()
                columns = [f.strip() for f in args["fields"].split(",")]
                definition["constraints"].append(
                    {"constraint_type": "UNIQUE", "columns": columns}
                )

        def get_datatype(coldef):
            dt = coldef["data_type"].upper()
            if dt == "CHARACTER":
                dt = "CHAR(%d)" % coldef["character_maximum_length"]
            elif dt == "CHARACTER VARYING":
                dt = "VARCHAR(%d)" % coldef["character_maximum_length"]
            elif dt == "DOUBLE PRECISION":
                dt = "FLOAT"

            if (coldef["column_default"] or "").startswith("nextval"):
                if "INT" not in dt:
                    raise NotImplementedError(dt)
                else:
                    dt = dt.replace("INT", "SERIAL")

            return dt

        # fix columns: list instead of dict
        for name, coldef in sorted(
            res["columns"].items(), key=lambda c: c[1]["ordinal_position"]
        ):
            col = {
                "name": name,
                "data_type": get_datatype(coldef),
                "is_nullable": coldef["is_nullable"],
            }
            if coldef.get("primary_key", False):
                col["primary_key"] = True
            if "foreign_key" in coldef:
                col["foreign_key"] = coldef["foreign_key"]
            definition["columns"].append(col)

        return definition

    def table_exists(self, table):
        """True or False

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

        """

        try:
            return self._table_exists(table)
        except OepTableNotFoundException:
            return False

    @check_exception("", OepTableNotFoundException)
    def _table_exists(self, table):
        """True or False

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

        """
        url = self._get_table_api_url(table=table)
        self._request("GET", url, 200)
        return True

    def get_metadata(self, table):
        """Returns metadata json

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

        """
        if not self.table_exists(table=table):
            raise OepTableNotFoundException
        url = self._get_table_api_url(table=table) + "meta/"
        res = self._request("GET", url, 200)
        return res

    @check_exception("not found", OepTableNotFoundException)
    def set_metadata(self, table, metadata):
        """write  metadata json, return accepted data from server

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

            metadata(object): json serializable object that follows the meta data specs
        """
        if not self.table_exists(table=table):
            raise OepTableNotFoundException
        url = self._get_table_api_url(table=table) + "meta/"
        metadata = self.validate_metadata(table, metadata)
        self._request("POST", url, 200, metadata)
        return self.get_metadata(table=table)

    def validate_metadata(self, table, data):
        if "id" not in data:
            data["id"] = table
        return data

    def advanced_session(self):
        return AdvancedApiSession(self)

    def count_rows(self, table):
        query = {
            "type": "select",
            "from": [{"type": "table", "table": table}],
            "fields": [
                {
                    "type": "label",
                    "element": {
                        "type": "function",
                        "function": "count",
                        "operands": {
                            "type": "grouping",
                            "grouping": [
                                {"type": "column", "column": "*", "is_literal": True}
                            ],
                        },
                    },
                    "label": "rowcount",
                }
            ],
        }
        with self.advanced_session() as sas:
            # not data yet
            res = sas._command("search", query)
            content = res["content"]
            description = content["description"]
            fieldnames = [f[0] for f in description]
            res = sas._command("cursor/fetch_one")
            content = res["content"]
            rec = dict(zip(fieldnames, content))
            rowcount = rec["rowcount"]
        return rowcount

    def publish_table(self, table: str, topic: str):
        """Publish table into topic"""
        url = self._get_table_api_url(table=table) + "move_publish/%s/" % topic
        return self._request("POST", url, 200)

    def unpublish_table(self, table: str):
        """Unpublish table"""
        url = self._get_table_api_url(table=table) + "unpublish"
        return self._request("POST", url, 200)

    def delete_from_table(self, table):
        """Delete all rows from table (without dropping it).

        Args:
            table(str): table name. Must be valid postgres table name,
                all lowercase, only letters, numbers and underscore

        """
        with self.advanced_session() as sas:
            sas.delete_from_table(table)

    def iter_tables(self):
        adv = AdvancedApiSession(self)  # no need to enter context
        url = adv.api_url + "get_table_names"

        tables = self._request(
            "post", url, jsondata={"query": {}}, expected_status=200
        )["content"]
        for table in tables:
            yield {"table": table}



# coding: utf-8
__version__ = "0.19.0"


from .advanced_api import AdvancedApiSession

# from .dialect import get_sqlalchemy_table
from .exceptions import (
    OepApiException,
    OepAuthenticationException,
    OepClientSideException,
    OepServerSideException,
    OepTableAlreadyExistsException,
    OepTableNotFoundException,
)
from .oep_client import (
    DEFAULT_API_VERSION,
    DEFAULT_BATCH_SIZE,
    DEFAULT_HOST,
    DEFAULT_INSERT_RETRIES,
    DEFAULT_PROTOCOL,
    TOKEN_ENV_VAR,
    OepClient,
)

__all__ = [
    "AdvancedApiSession",
    # "get_sqlalchemy_table",
    "OepApiException",
    "OepAuthenticationException",
    "OepClientSideException",
    "OepServerSideException",
    "OepTableAlreadyExistsException",
    "OepTableNotFoundException",
    "DEFAULT_API_VERSION",
    "DEFAULT_BATCH_SIZE",
    "DEFAULT_HOST",
    "DEFAULT_INSERT_RETRIES",
    "DEFAULT_PROTOCOL",
    "TOKEN_ENV_VAR",
    "OepClient",
]



"""Command line script for OepClient"""

__version__ = "0.15.0"

import json
import logging
import re
from copy import deepcopy
from tempfile import NamedTemporaryFile
from urllib.parse import urlsplit

import pandas as pd
import requests
from numpy import nan

from oep_client.exceptions import OepClientSideException

DEFAULT_INDENT = 2


def read_json(filepath, encoding):

    filepath = make_local(filepath)

    with open(filepath, encoding=encoding) as file:
        return json.load(file)


def write_json(data, filepath, encoding):
    with open(filepath, "w", encoding=encoding) as file:
        json.dump(data, file, indent=DEFAULT_INDENT, ensure_ascii=False)


def read_metadata_json(filepath, encoding):
    # TODO: valdiation?
    return read_json(filepath, encoding)


def get_schema_definition_from_metadata(meta):
    definition = meta["resources"][0]["schema"]
    definition = fix_table_definition(definition)
    return definition


def fix_name(name):
    name_new = re.sub("[^a-z0-9_]+", "_", name.lower())
    if name_new != name:
        logging.warning('Changed name "%s" to "%s"', name, name_new)
    return name_new


def dataframe_to_records(df):
    columns = [fix_name(n) for n in df.columns]
    # replace nan
    df = df.replace({nan: None})
    data = df.values.tolist()
    data = [dict(zip(columns, row)) for row in data]
    return data


def records_to_dataframe(records):
    df = pd.DataFrame(records)
    return df


def make_local(filepath_or_url: str) -> str:
    # if filepath is url: download to tempfile
    if not re.match("^http[s]?://", filepath_or_url):
        return filepath_or_url
    suffix = urlsplit(filepath_or_url).path.split("/")[-1]
    suffix = re.sub("[^a-z0-9.]", "_", suffix.lower())  # replace non word chars
    with NamedTemporaryFile(mode="wb", delete=False, suffix="_" + suffix) as file:
        logging.debug(f"Downloading {filepath_or_url} => {file.name}")
        resp = requests.get(filepath_or_url)
        resp.raise_for_status()
        file.write(resp.content)
        return file.name


def read_dataframe(filepath, **kwargs):

    filepath = make_local(filepath)

    if filepath.endswith(".json"):
        df = pd.read_json(filepath)
    elif filepath.endswith(".csv"):
        df = pd.read_csv(
            filepath, encoding=kwargs.get("encoding"), sep=kwargs.get("delimiter")
        )
    elif filepath.endswith(".xlsx"):
        # pd.read_excel default for sheet_name = 0 (first sheet)
        sheet = kwargs.get("sheet", 0)
        df = pd.read_excel(filepath, sheet)
    else:
        raise OepClientSideException("Unsupported filetype: %s" % filepath)
    return df


def write_dataframe(df, filepath, **kwargs):
    if filepath.endswith(".json"):
        df.to_json(filepath, orient="records", indent=DEFAULT_INDENT, force_ascii=False)
    elif filepath.endswith(".csv"):
        df.to_csv(
            filepath,
            encoding=kwargs.get("encoding"),
            sep=kwargs.get("delimiter"),
            index=False,
        )
    elif filepath.endswith(".xlsx"):
        sheet = kwargs.get("sheet")
        if not sheet:
            raise OepClientSideException("Must specify sheet when reading excel files")
        df.to_excel(filepath, sheet, index=False)
    elif filepath == "-":
        # stdout
        s = df.to_string(index=False)
        print(s)
    else:
        raise OepClientSideException("Unsupported filetype: %s" % filepath)
    return df


def fix_table_definition(definition):
    definition = deepcopy(definition)
    if "fields" in definition:
        definition["columns"] = definition.pop("fields")
    definition["columns"] = [fix_column_definition(c) for c in definition["columns"]]
    return definition


def fix_column_definition(definition):
    definition = deepcopy(definition)
    definition["data_type"] = definition.get("data_type") or definition.pop("type")
    return definition



# coding: utf-8
import json
import logging
import os
import random
import unittest

from . import TOKEN_ENV_VAR, OepClient
from .exceptions import OepClientSideException

MAX_TRIES_FIND_RANDOM_TEST_TABLE = 10


TEST_TABLE_DEFINITION = {
    "columns": [
        # {
        #     "name": "id",
        #     "type": "integer",
        #     "data_type": "bigserial",
        #     "is_nullable": False,
        #     "primary_key": True,
        # },
        {
            "name": "field1",
            "type": "string",
            "data_type": "varchar(128)",
            "is_nullable": False,
        },
        {"name": "field2", "type": "integer", "is_nullable": True},
    ],
    "constraints": [{"constraint_type": "UNIQUE", "columns": ["field1"]}],
}


class TestRoundtrip(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        token = os.environ.get(TOKEN_ENV_VAR)
        if not token:
            raise Exception(
                "In order to run the test, you must set the environment variable %s"
                % TOKEN_ENV_VAR
            )

        cls.client = OepClient(token=token)

    def test_roundtrip(self, client=None):
        """
        * create table
        * upload data
        * download data
        * delete table
        """

        client = client or self.client

        # generate test data
        test_data = [
            {"field1": "test unicode 😀"},
            {"field1": "k1", "field2": 1},
            {"field1": "k2", "field2": 2},
            {"field1": "k3", "field2": 3},
        ]

        # create a random test table name that does not exist
        tries_left = MAX_TRIES_FIND_RANDOM_TEST_TABLE
        while tries_left:
            table_name = "test_table_%s" % random.randint(0, 1000000000)
            if not client.table_exists(table_name):
                break
            if not tries_left:
                raise Exception(
                    "Could not create a random test table name after %d tries"
                    % MAX_TRIES_FIND_RANDOM_TEST_TABLE
                )

        tdef = client.create_table(table_name, TEST_TABLE_DEFINITION, is_sandbox=True)
        logging.info(tdef)

        rcount = client.insert_into_table(table_name, test_data)

        # insert second time should fail because unique constraint
        self.assertRaises(
            OepClientSideException,
            client.insert_into_table,
            table_name,
            test_data,
        )

        self.assertEqual(rcount, len(test_data))
        data = client.select_from_table(table_name)
        # remove generated id columns
        for row in data:
            del row["id"]

        # test equality
        self.assertEqual(
            json.dumps(test_data, sort_keys=True), json.dumps(data, sort_keys=True)
        )

        # also test where
        data_partial = client.select_from_table(
            table_name, where=["field2>1", "field2<3"]
        )
        self.assertEqual(len(data_partial), 1)
        self.assertEqual(data_partial[0]["field2"], 2)

        # test count_rows
        self.assertEqual(client.count_rows(table_name), len(test_data))

        # test publish: must have license
        metadata = {
            "resources": [{"licenses": [{"name": "CC0"}]}],
            "metaMetadata": {"metadataVersion": "OEMetadata-2.0.0"},
        }
        client.set_metadata(table_name, metadata)
        client.publish_table(table_name, "scenario")
        client.unpublish_table(table_name)

        # delete data
        client.delete_from_table(table_name)
        self.assertEqual(client.count_rows(table_name), 0)

        # drop table
        client.drop_table(table_name)


class TestUtils(unittest.TestCase):
    """functions without token auth"""

    @classmethod
    def setUpClass(cls):
        cls.client = OepClient()

    def test_iter_tables(self):
        self.assertTrue(
            all(set(["table"]) == set(x) for x in self.client.iter_tables())
        )
